#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#define ALLOW_EMPTY_CREDENTIALS 1
#define LCMAPS_GSI_MODE 1
#include <lcmaps.h>
#include <lcmaps_cred_data.h>

/* proxy.c */
gss_cred_id_t read_proxy(const char* filename);
void free_proxy(gss_cred_id_t cred);
/* cred2dn.c */
char * gss_cred_to_dn(gss_cred_id_t globus_cred);

#define usage(x) fprintf(stderr, "Usage: %s [-v] [-D DN] -P <proxyfile> [-G] -f <configfile> [-d lcmaps_dir] policy...\n", x)

int main(int argc, char *argv[])
{
    int opt;
    int verbose = 0;
    int nogroup = 0;
    int npols;
    char **policynames;
    int fail = 0;
    const char *credfile = NULL;
    const char *conffile = NULL;
    const char *lcmapsdir = NULL;
    char *creddn = NULL;

    FILE *lcmaps_log;
    gss_cred_id_t cred;
    char *username = NULL;
    uid_t *uids;
    int n_uids;
    gid_t *gids;
    int n_gids;

    struct passwd *pwd;
    struct group *grp;

    while ((opt = getopt(argc, argv, "D:P:f:d:vG")) != -1) {
        switch (opt) {
        case 'v':
            verbose = 1;
            break;
        case 'G':
            nogroup = 1;
            break;
        case 'D':
            creddn = optarg;
            break;
        case 'P':
            credfile = optarg;
            break;
        case 'f':
            conffile = optarg;
            break;
        case 'd':
            lcmapsdir = optarg;
            break;
        default: /* '?' */
            usage(argv[0]);
            exit(EXIT_FAILURE);
        }
    }
    if(!credfile) {
        fprintf(stderr, "%s: No credential file supplied!\n", argv[0]);
        usage(argv[0]);
        fail = 1;
        goto exit0;
    }
    if(!conffile) {
        fprintf(stderr, "%s: No LCMAPS configuration file supplied!\n", argv[0]);
        usage(argv[0]);
        fail = 1;
        goto exit0;
    }
    if (optind >= argc) {
        fprintf(stderr, "%s: Expected argument after options\n", argv[0]);
        usage(argv[0]);
        fail = 1;
        goto exit0;
    }
    policynames = argv + optind;
    npols = argc - optind;

    setenv("LCMAPS_DB_FILE", conffile, 1);
    if(lcmapsdir)
        setenv("LCMAPS_DIR", lcmapsdir, 1);

    cred = read_proxy(credfile);
    if(!cred) {
        fprintf(stderr, "%s: Failed to load proxy from file: %s\n", argv[0], credfile);
        fail = 1;
        goto exit0;
    }

    if(!creddn) {
        if(verbose)
            fprintf(stderr, "%s: Subject DN not supplied, trying to retrieve from proxy\n", argv[0]);
        creddn = gss_cred_to_dn(cred);
    }

    lcmaps_log = verbose?fdopen(STDERR_FILENO,"a"):fopen("/dev/null","a");
    if(lcmaps_init(lcmaps_log) != 0) {
        fprintf(stderr, "%s: Failed initialize LCMAPS\n", argv[0]);
        fail = 1;
        goto exit2;
    }
    if(lcmaps_run_and_return_username(creddn, cred, (char*)"", &username, npols, policynames) != 0) {
        fprintf(stderr, "%s: Failed to get user mapping\n", argv[0]);
        fail = 1;
        goto exit3;
    }
    if(username == NULL) {
        fprintf(stderr, "%s: LCMAPS succeeded but returned no username\n", argv[0]);
        fail = 1;
        goto exit3;
    }

    if(((uids = getCredentialData(UID, &n_uids)) == NULL) || (n_uids < 1)) {
        fprintf(stderr, "%s: LCMAPS succeeded but did not returned any primary UIDs\n", argv[0]);
        fail = 1;
        goto exit3;
    }
    if(uids[0] == 0) {
        fprintf(stderr, "%s: LCMAPS succeeded but mapped to root account!\n", argv[0]);
        fail = 1;
        goto exit3;
    }
    if((pwd = getpwuid(uids[0])) == NULL ) {
        fprintf(stderr, "%s: LCMAPS could not find the username related to uid: %d\n", argv[0], uids[0]);
        fail = 1;
        goto exit3;
    }
    printf("%s", pwd->pw_name);

    if(!nogroup) {
        if(((gids = getCredentialData(PRI_GID, &n_gids)) == NULL) || (n_gids < 1)) {
            if(verbose)
                fprintf(stderr, "%s: LCMAPS succeeded but did not returned any primary GIDs\n", argv[0]);
            goto exit3;
        }
        if(gids[0] == 0) {
            fprintf(stderr, "%s: LCMAPS succeeded but mapped to root group!\n", argv[0]);
            fail = 1;
            goto exit3;
        }
        if((grp = getgrgid(gids[0])) == NULL ) {
            if(verbose)
                fprintf(stderr, "%s: LCMAPS could not find the group name related to gid: %d\n", argv[0], gids[0]);
            goto exit3;
        }
        printf(":%s", grp->gr_name);
    }

exit3:
    lcmaps_term();
exit2:
    fclose(lcmaps_log);
exit1:
    free_proxy(cred);
exit0:
    exit(fail?EXIT_FAILURE:EXIT_SUCCESS);
}
